import { Component, ViewChild } from '@angular/core';

import {
    acSection, cSnTrapez, ParamsSectionTrapez, cSnRectang, ParamsSectionRectang, cSnCirc,
    ParamsSectionCirc, cSnPuiss, ParamsSectionPuiss, Result
} from 'jalhyd';

import { CalcCanvasComponent } from '../canvas/canvas.component';

@Component({
    selector: 'section-canvas',
    template: `<calc-canvas #calcCanvas
    [width] = '_size'
    [height] = '_size'>
    </calc-canvas>
    `
})
export class SectionCanvasComponent {
    /**
     * taille (pixels) du canvas (c'est un carré)
     */
    private _size: number = 400;

    /**
     * marges gauche/droite pour le texte (pixels)
     */
    private _textMargin: number = 90;

    /**
     * marge basse (pixels)
     */
    private _bottomMargin: number = 5;

    /**
     * facteurs d'échelle (coordonnées en m <-> coordonnées en pixels)
     */
    private _scaleX: number;
    private _scaleY: number;

    private _section: acSection;

    // tirants
    private _levels: Object[] = [];

    @ViewChild("calcCanvas")
    private _calcCanvas: CalcCanvasComponent;

    public set section(s: acSection) {
        this._section = s;
        if (this._section != undefined)
            this.draw();
    }

    public reset() {
        this._section = undefined;
        this._levels = [];
        this._calcCanvas.clear();
    }

    private computeScale(maxWidth: number, maxHeight: number) {
        this._scaleX = (this._size - 2 * this._textMargin) / maxWidth;
        this._scaleY = (this._size - this._bottomMargin) / maxHeight;
    }

    /**
     * convertit une abscisse en m en pixels
     */
    private Xm2pix(x: number) {
        return this._textMargin + x * this._scaleX
    }

    /**
     * convertit une ordonnée en m en pixels
     */
    private Ym2pix(y: number) {
        return this._size - this._bottomMargin - y * this._scaleY
    }

    public addLevel(val: number, label: string, rgb: {}) {
        this._levels.push({ val, label, rgb });
    }

    // max des niveaux à représenter
    private getMaxLevel(): number {
        // max de la cote de berge et des niveaux (qui sont triés)
        return Math.max(this._section.prms.YB.v, this._levels[this._levels.length - 1]["val"]);
    }

    /**
     * dessin des pointillés au dessus de la berge
     * xmin,xmax : position gauche/droite
     * yb : cote de berge
     * maxHeight : valeur maxi des cotes
     */
    private drawTopDashLines(xmin: number, xmax: number, yb: number, maxHeight: number) {
        this._calcCanvas.setLineWidth(2);
        this._calcCanvas.setLineDash([5]);
        this._calcCanvas.setStrokeColor(128, 128, 128);
        this.drawSectionLine(xmin, yb, xmin, maxHeight);
        this.drawSectionLine(xmax, yb, xmax, maxHeight);
    }

    private drawSectionTrapez(): number {
        let sect: cSnTrapez = <cSnTrapez>this._section;
        let prms: ParamsSectionTrapez = <ParamsSectionTrapez>sect.prms;

        // cote de berge
        let yb: number = prms.YB.v;

        // largeur de la partie pentue
        let lp: number = prms.Fruit.v * prms.YB.v;

        // largeur totale de la section
        let maxWidth: number = lp * 2 + prms.LargeurFond.v;

        // hauteur totale de la section
        let maxHeight: number = this.getMaxLevel();
        maxHeight *= 1.1;

        this.computeScale(maxWidth, maxHeight);

        // dessin de la section

        this._calcCanvas.setStrokeColor(0, 0, 0);
        this._calcCanvas.setLineWidth(5);
        this.drawSectionLine(0, yb, lp, 0);
        this.drawSectionLine(lp, 0, lp + prms.LargeurFond.v, 0);
        this.drawSectionLine(lp + prms.LargeurFond.v, 0, maxWidth, yb);

        // pointillés du haut

        this.drawTopDashLines(0, maxWidth, yb, maxHeight);

        return maxWidth;
    }

    private drawSectionRect() {
        let sect: cSnRectang = <cSnRectang>this._section;
        let prms: ParamsSectionRectang = <ParamsSectionRectang>sect.prms;

        // cote de berge
        let yb: number = prms.YB.v;

        // largeur totale de la section
        let maxWidth: number = prms.LargeurBerge.v;

        // hauteur totale de la section
        let maxHeight: number = this.getMaxLevel();
        maxHeight *= 1.1;

        this.computeScale(maxWidth, maxHeight);

        // dessin de la section

        this._calcCanvas.setStrokeColor(0, 0, 0);
        this._calcCanvas.setLineWidth(5);
        this.drawSectionLine(0, yb, 0, 0);
        this.drawSectionLine(0, 0, prms.LargeurBerge.v, 0);
        this.drawSectionLine(prms.LargeurBerge.v, 0, maxWidth, yb);

        // pointillés du haut

        this.drawTopDashLines(0, maxWidth, yb, maxHeight);

        return maxWidth;
    }

    private drawSectionCirc() {
        let sect: cSnCirc = <cSnCirc>this._section;
        let prms: ParamsSectionCirc = <ParamsSectionCirc>sect.prms;

        // cote de berge
        let yb: number = prms.YB.v;

        // diamètre, rayon
        let D: number = prms.D.v;
        let r: number = D / 2;

        try {
            // largeur au miroir
            let B: Result = sect.Calc("B", yb);
            if (!B.ok)
                throw B;

            // largeur totale de la section
            let maxWidth: number = yb < r ? B.vCalc : D;

            // hauteur totale de la section
            let maxHeight: number = this.getMaxLevel();
            maxHeight *= 1.1;

            this.computeScale(maxWidth, maxHeight);

            // dessin de la section

            this._calcCanvas.setStrokeColor(0, 0, 0);
            this._calcCanvas.setLineWidth(5);

            let wx: number = maxWidth / 2;
            let alpha: Result = sect.Calc("Alpha", yb);
            if (!alpha.ok)
                throw alpha;

            let s: number = Math.PI / 2 - alpha.vCalc;
            let e: number = Math.PI / 2 + alpha.vCalc;
            this.drawSectionEllipse(wx, r, r, r, 0, s, e);

            // pointillés du haut

            let w: number = yb > r ? (D - B.vCalc) / 2 : 0;
            this.drawTopDashLines(w, maxWidth - w, yb, maxHeight);

            return maxWidth;
        }
        catch (e) {
            let r: Result = e as Result;
            this.drawText("error : " + r.log.toString(), 0, 0);
        }
    }

    private drawSectionPara() {
        let sect: cSnPuiss = <cSnPuiss>this._section;
        let prms: ParamsSectionPuiss = <ParamsSectionPuiss>sect.prms;

        // cote de berge
        let yb: number = prms.YB.v;

        // largeur au miroir
        let B: number = prms.LargeurBerge.v;

        // largeur totale de la section
        let maxWidth: number = B;

        // hauteur totale de la section
        let maxHeight: number = this.getMaxLevel();
        maxHeight *= 1.1;

        this.computeScale(maxWidth, maxHeight);

        // contour de la section

        this._calcCanvas.setStrokeColor(0, 0, 0);
        this._calcCanvas.setLineWidth(5);
        let k: number = prms.k.v;
        let lambda: number = B / Math.pow(yb, k);
        let inv_k: number = 1 / k;

        let n = 20;
        this.beginPath();
        for (let x: number = -B / 2; x <= B / 2; x += B / n) {
            let y: number = Math.pow(Math.abs(x) * 2 / lambda, inv_k);
            this.lineTo(x + B / 2, y);
        }
        this.closePath();

        // pointillés du haut
        this.drawTopDashLines(0, maxWidth, yb, maxHeight);

        return maxWidth;
    }

    private beginPath() {
        this._calcCanvas.context2d.beginPath();
    }

    private lineTo(x: number, y: number) {
        this._calcCanvas.context2d.lineTo(this.Xm2pix(x), this.Ym2pix(y));
    }

    private closePath() {
        this._calcCanvas.context2d.stroke();
    }

    private drawSectionEllipse(x: number, y: number, rX: number, rY: number, rot: number, start: number, end: number) {
        this._calcCanvas.drawEllipse(this.Xm2pix(x), this.Ym2pix(y), rX * this._scaleX, rY * this._scaleY, rot, start, end);
    }

    private drawText(s: string, x: number, y: number, align: string = undefined) {
        this._calcCanvas.fillText(s, this.Xm2pix(x), this.Ym2pix(y), align);
    }

    private drawLine(x1: number, y1: number, x2: number, y2: number) {
        this._calcCanvas.drawLine(x1 * this._scaleX, this._size - y1 * this._scaleY, x2 * this._scaleX, this._size - y2 * this._scaleY);
    }

    private drawSectionLine(x1: number, y1: number, x2: number, y2: number) {
        this._calcCanvas.drawLine(this.Xm2pix(x1), this.Ym2pix(y1), this.Xm2pix(x2), this.Ym2pix(y2));
    }

    /**
     * dessin de la section
     * @returns largeur de la section (en m)
     */
    private drawSection(): number {
        if (this._section instanceof cSnTrapez)
            return this.drawSectionTrapez();
        if (this._section instanceof cSnRectang)
            return this.drawSectionRect();
        if (this._section instanceof cSnCirc)
            return this.drawSectionCirc();
        if (this._section instanceof cSnPuiss)
            return this.drawSectionPara();
        throw "SectionCanvasComponent.drawSection() : type de section non pris en charge";
    }

    private sortLevels() {
        this._levels.sort((a, b) => {
            if (a["val"] < b["val"])
                return -1;
            if (a["val"] > b["val"])
                return 1;
            return 0;
        })
    }

    private drawLevels(maxWidth: number) {
        let left = true;

        this._calcCanvas.resetLineDash();
        this._calcCanvas.setLineWidth(1);
        this._calcCanvas.setFont("12px sans- serif")
        for (let l of this._levels) {
            let y = l["val"];
            let col = l["rgb"];
            this._calcCanvas.setStrokeColor(col["r"], col["g"], col["b"]);
            this.drawSectionLine(0, y, maxWidth, y);

            this._calcCanvas.setFillColor(col["r"], col["g"], col["b"]);
            if (left)
                this.drawText(l["label"], -0.1, y, "right");
            else
                this.drawText(l["label"], maxWidth + 0.1, y, "left");
            left = !left;
        }
    }

    // contour du canvas
    private drawFrame() {
        this._calcCanvas.clear();
        this._calcCanvas.resetLineDash();
        this._calcCanvas.setStrokeColor(128, 128, 128);
        this._calcCanvas.drawRect(0, 0, this._calcCanvas.width, this._calcCanvas.height);
    }

    private draw() {
        this.sortLevels();
        this.drawFrame();
        let maxWidth = this.drawSection();
        this.drawLevels(maxWidth);
    }
}
