import { CalculatorType, ComputeNodeType, Nub, ParamDefinition, SessionNub, Props, NubFactory, Observer } from "jalhyd";

import { FormulaireElement } from "../formulaire-element";
import { NgParameter, ParamRadioConfig } from "../ngparam";
import { Field } from "../field";
import { StringMap } from "../../stringmap";
import { FormulaireNode } from "../formulaire-node";
import { ServiceFactory } from "../../services/service-factory";
import { ParamService } from "../../services/param/param.service";
import { FieldSet } from "../fieldset";
import { FieldsetContainer } from "../fieldset-container";
import { SelectField } from "../select-field";
import { DeepFieldsetIterator } from "../form-iterator/deep-fieldset-iterator";
import { DeepFormulaireElementIterator } from "../form-iterator/deep-element-iterator";
import { TopFormulaireElementIterator } from "../form-iterator/top-element-iterator";
import { CalculatorResults } from "../../results/calculator-results";
import { FieldsetTemplate } from "../fieldset-template";

/**
 * classe de base pour tous les formulaires
 */
export abstract class FormulaireDefinition extends FormulaireNode implements Observer {
    /**
     * nom de la calculette
     */
    private _calculatorName: string;

    /**
     * SessionNub courant
     */
    protected _currentSessionNub: SessionNub;

    /**
     * fichier de configuration
     */
    private _jsonConfig: {};

    /**
     * aide en ligne
     */
    private _helpText: string;

    protected _paramService: ParamService;

    constructor() {
        super(undefined);
        this._paramService = ServiceFactory.instance.paramService;
    }

    public get calculatorType(): CalculatorType {
        const props = this._currentSessionNub === undefined ? this.defaultProperties : this._currentSessionNub.properties.props;
        return props["calcType"];
    }

    public get nodeType(): ComputeNodeType {
        const props = this._currentSessionNub === undefined ? this.defaultProperties : this._currentSessionNub.properties.props;
        return props["nodeType"];
    }

    public get calculatorName() {
        return this._calculatorName;
    }

    public set calculatorName(name: string) {
        this._calculatorName = name;
    }

    public get jsonConfig(): {} {
        return this._jsonConfig;
    }

    protected abstract get defaultProperties(): {};

    public initSessionNub(props?: {}) {
        this._currentSessionNub = this.createSessionNub(props === undefined ? this.defaultProperties : props);
    }

    public get currentSessionNub(): SessionNub {
        return this._currentSessionNub;
    }

    private findNub(params: Props | {}) {
        return this._paramService.findSessionNub(params);
    }

    protected createSessionNub(params: Props | {}): SessionNub {
        return this._paramService.createSessionNub(params);
    }

    /**
     * @return le Nub associé à un type de noeud donné, le crée si nécessaire
     * @param params paramètres de contexte de création du nub
     */
    protected getSessionNub(params: Props | {}): SessionNub {
        let res = this.findNub(params);
        if (!res)
            res = this.createSessionNub(params);
        return res;
    }

    protected replaceCurrentSessionNub(params: Props) {
        this._currentSessionNub = this._paramService.replaceSessionNub(this._currentSessionNub, params);
    }

    protected replaceSessionNub(sn: SessionNub, params: Props): SessionNub {
        return this._paramService.replaceSessionNub(sn, params);
    }

    /**
     * @return le paramètre d'un type de noeud
     * @param symbol symbole du paramètre
     * @param params paramètres de contexte de création du nub
     */
    protected getNubParamFromSymbol(symbol: string, params: Props | {}): ParamDefinition {
        let sessionNub: SessionNub = this.getSessionNub(params);

        if (sessionNub)
            return sessionNub.nub.getParameter(symbol);

        throw new Error(`FormulaireDefinition.getNubParamFromSymbol() : pas de Nub trouvé pour ${params}`)
    }

    protected deleteSessionNub(sn: SessionNub) {
        this._paramService.deleteSessionNub(sn);
    }

    protected initParse() {
    }

    protected parseOptions(json: {}) {
    }

    protected completeParse(json: {}) {
    }

    public getOption(json: {}, option: string): string {
        if (json["type"] === "options")
            return json[option];

        return undefined;
    }

    public afterParseFieldset(fs: FieldSet) {
    }

    public createFieldset(parent: FormulaireNode, json: {}, data?: {}): FieldSet {
        const res: FieldSet = new FieldSet(parent);
        res.setSessionNub(this._currentSessionNub, false);
        this.kids.push(res);
        return res;
    }

    public moveFieldsetUp(fs: FieldSet) {
    }

    public moveFieldsetDown(fs: FieldSet) {
    }

    public removeFieldset(fs: FieldSet) {
    }

    private parse_fieldset(json: {}) {
        const fs = this.createFieldset(this, json);
        fs.parseConfig(json);
        this.afterParseFieldset(fs);
    }

    private parse_template_container(json: {}, templates: any[]) {
        const fsc: FieldsetContainer = new FieldsetContainer(this);
        fsc.parseConfig(json, templates);
        this.formElements.push(fsc);
    }

    public parseDependencies(json: {}) {
        for (let conf_index in json) {
            const conf = json[conf_index];
            const type: string = conf["type"];

            switch (type) {
                // field set
                case "fieldset":
                case "template_container":
                    for (const k of this.kids)
                        if (k.id == conf["id"]) {
                            k.parseDependencies(conf);
                            break;
                        }
                    break;

                case "fieldset_template":
                    for (const k of this.kids)
                        if (k instanceof FieldsetContainer)
                            k.parseDependencies(conf);
                    break;
            }
        }
    }

    public parseConfig(json: {}) {
        this._jsonConfig = json;

        this.initParse();

        // analyse des options globales
        // il est utile de le faire avant le reste pour les calculettes utilisant des sections (id des selects type de section/variable à calculer)

        for (let conf_index in json) {
            const conf = json[conf_index];
            const type: string = conf["type"];

            if (type === "options") {
                this.parseOptions(conf);
                break;
            }
        }

        // analyse des éléments du formulaire

        const templates: any[] = [];

        for (let conf_index in json) {
            const conf = json[conf_index];
            const type: string = conf["type"];

            switch (type) {
                case "fieldset":
                    this.parse_fieldset(conf);
                    break;

                case "fieldset_template":
                    templates.push(conf);
                    break;

                // options globales
                case "options":
                    break;

                case "template_container":
                    this.parse_template_container(conf, templates);
                    break;

                default:
                    throw new Error(`type d'objet de calculette ${type} non pris en charge`);
            }
        }

        this.completeParse(json);

        // console.log("-----");

        // for (const n of NubFactory.getInstance().sessionNubIterator) {
        //     console.log(n.nub);
        //     for (const p of n.nub.parameterIterator)
        //         console.log(`${p.symbol} uid  ${p.uid} props ${n.properties} mode ${p.valueMode} val ${p.uncheckedValue}`);
        // }

        // logObject(this._fieldSets, "fieldsets");
        // logObject(this._dependencies, "dependences");

        this.parseDependencies(json);
    }

    public hasParameter(symbol: string): boolean {
        for (const p of this.allFormElements) {
            if (p instanceof NgParameter)
                if (p.symbol === symbol)
                    return true;
        }
        return false;
    }

    public getParamFromSymbol(symbol: string): NgParameter {
        for (const p of this.allFormElements) {
            if (p instanceof NgParameter)
                if (p.symbol === symbol)
                    return p;
        }
        return undefined;
    }

    public getDisplayedParamFromState(st: ParamRadioConfig): NgParameter {
        for (const p of this.allFormElements) {
            if (p.isDisplayed && p instanceof NgParameter)
                if (p.radioState == st)
                    return p;
        }
        return undefined;
    }

    public getDisplayedParamListFromState(st: ParamRadioConfig): NgParameter[] {
        const res = [];
        for (const p of this.allFormElements)
            if (p.isDisplayed && p instanceof NgParameter && p.radioState == st)
                res.push(p)
        return res;
    }

    public getFieldById(id: string): Field {
        let res = this.getFormulaireNodeById(id);
        if (res instanceof Field)
            return res;
        return undefined;
    }

    public getParameterValue(symbol: string): number {
        for (const fs of this.allFieldsets) {
            const p = fs.getNodeParameter(symbol);
            if (p != undefined)
                switch (p.radioState) {
                    case ParamRadioConfig.FIX:
                        return p.getValue();

                    case ParamRadioConfig.VAR:
                    case ParamRadioConfig.CAL:
                        return undefined;
                }
        }

        throw new Error(`Formulaire.getNodeParameterValue() : pas de paramètre ${symbol} trouvé`);
    }

    private notifyReset() {
        this.notifyObservers({
            "action": "resetForm"
        }, this);
    }

    /**
     * réinitialisation du formulaire suite à un changement d'une valeur, d'une option, ... :
     * effacement des résultats, application des dépendances, ...
     */
    public reset() {
        this.resetResults();
        this.applyDependencies();

        // prévenir les composants qu'il faut détecter les changements
        this.notifyReset();
    }

    /**
     * retourne la valeur actuellement sélectionnée d'un SelectField
     * @param selectFieldId id du SelectField
     * @returns valeur courante du select sans le préfixe
     */
    public getSelectedValue(selectFieldId: string): any {
        let select: SelectField = <SelectField>this.getFieldById(selectFieldId);
        return select.getValue().value;
    }

    /**
     * retourne le label actuellement sélectionnée d'un SelectField
     * @param selectFieldId id du SelectField
     * @returns label courant du select
     */
    public getSelectedLabel(selectFieldId: string): string {
        let select: SelectField = <SelectField>this.getFieldById(selectFieldId);
        return select.getValue().label;
    }

    public get formElements(): FormulaireElement[] {
        return this.kids as FormulaireElement[];
    }

    public applyDependencies() {
        for (const fe of this.topFormElements)
            fe.applyDependencies(this);
    }

    public abstract resetResults();
    public abstract doCompute();
    public abstract get hasResults(): boolean;
    public abstract get results(): CalculatorResults[];

    public updateLocalisation(localisation: StringMap) {
        for (let fe of this.topFormElements)
            fe.updateLocalisation(localisation);

        if (this.hasResults)
            this.doCompute(); // pour mettre à jour la langue
    }


    public updateHelp() {
        ServiceFactory.instance.helpService.loadFormHelp(this, ServiceFactory.instance.internationalisationService.currentLanguage);
    }

    public get helpText(): string {
        return this._helpText;
    }

    public set helpText(h: string) {
        this._helpText = h;
    }

    public requestHelp() {
        this.notifyObservers({ "action": "requestHelp" }, this);
    }

    public isDisplayed(id: string) {
        return (<FormulaireElement>this.getFormulaireNodeById(id)).isDisplayed;
    }

    public get isValid(): boolean {
        let res: boolean = true;
        for (let fs of this.allFieldsets)
            if (fs.isDisplayed)
                res = res && fs.isValid;
        return res;
    }

    /**
     * gestion d'un clic sur les radios
     */
    public onRadioClick(info: any) {
        this.reset();
    }

    /**
     * itère sur tous les FieldSet
     */
    public get allFieldsets(): IterableIterator<FieldSet> {
        return new DeepFieldsetIterator(this);
    }

    /**
     * itère sur tous les FormulaireElement
     */
    public get allFormElements(): IterableIterator<FormulaireElement> {
        return new DeepFormulaireElementIterator(this);
    }

    /**
     * itère sur tous les FormulaireElement de 1er niveau
    */
    private get topFormElements(): IterableIterator<FormulaireElement> {
        return new TopFormulaireElementIterator(this);
    }

    /**
     * sérialisation en JSON
     */
    public JSONserialise(): {} {
        let res = {};
        res["id"] = this.calculatorName;
        res["uid"] = this.uid;
        res["props"] = this._currentSessionNub.properties.props;
        res["elements"] = this.serialiseKids();
        return { "form": res };
    }

    private deserialiseFieldset(elements: {}): FieldSet {
        const res: FieldSet = this.getFormulaireNodeById(elements["id"]) as FieldSet;
        res.deserialiseJSON(elements);
        return res;
    }

    private deserialiseFieldsetContainer(elements: {}): FieldsetContainer {
        const res: FieldsetContainer = this.getFormulaireNodeById(elements["id"]) as FieldsetContainer;
        res.deserialiseJSON(elements);
        return res;
    }

    private deserialiseElement(element: {}) {
        const keys = Object.keys(element);
        if (keys.length !== 1)
            throw new Error(`session file : invalid form object '${element}'`);

        switch (keys[0]) {
            case "fieldset":
                this.deserialiseFieldset(element[keys[0]]);
                break;

            case "fieldset_container":
                this.deserialiseFieldsetContainer(element[keys[0]]);
                break;

            default:
                throw new Error(`session file : invalid key '${keys[0]}' in form object`);
        }
    }

    /**
     * désérialisation depuis JSON
     */
    public deserialiseJSON(elements: {}) {
        for (const k in elements) {
            switch (k) {
                case "id":
                    this._calculatorName = elements[k];
                    break;

                case "uid":
                case "props":
                    break;

                case "elements":
                    for (const e of elements[k])
                        this.deserialiseElement(e);
                    break;

                default:
                    throw new Error(`session file : invalid key '${k}' in form object`);
            }
        }
    }

    //  interface Observer

    public update(sender: any, data: any) {
    }
}
