import { FormulaireNode } from "./formulaire-node"
import { StringMap } from "../stringmap";
import { Dependency } from "./dependency/dependency";
import { DependencyCondition, DependencyConditionType } from "./dependency/dependency-condition";
import { ValueDependency } from "./dependency/value-dependency";
import { ValueDependencyCondition } from "./dependency/value-dependency-condition";
import { ExistenceDependency } from "./dependency/existence-dependency";
import { FormulaireDefinition } from "./definition/form-definition";
import { DeepFormulaireElementIterator } from "./form-iterator/deep-element-iterator";

/** 
 * élément (enfant) du formulaire : fieldset, input, container, ...
 */
export abstract class FormulaireElement extends FormulaireNode {
    /**
     * affiché ?
     */
    private _isDisplayed: boolean;

    /**
     * étiquette affichée dans l'UI
     */
    private _label: string;

    /**
     * n° d'ordre ajouté à l'étiquette
     */
    private _labelNumber: number;

    protected _dependencies: Dependency[] = [];

    constructor(parent: FormulaireNode) {
        super(parent);
        this._isDisplayed = true;
    }

    get isDisplayed(): boolean {
        return this._isDisplayed;
    }

    set isDisplayed(b: boolean) {
        this._isDisplayed = b;
        for (const k of this.getKids())
            k.isDisplayed = b;
    }

    get label(): string {
        let res = this._label;
        if (res && this._labelNumber)
            res += " n°" + String(this._labelNumber);
        return res;
    }

    public set labelNumber(n: number) {
        this._labelNumber = n;
    }

    public getKids(): FormulaireElement[] {
        return super.kids as FormulaireElement[];
    }

    /**
     * analyse les dépendances d'existence
     * @param json configuration de la dépendance
     */
    private parse_existence_dependencies(json: {}) {
        for (let di in json) {
            let d = json[di];
            let rv = d["refvalue"];
            if (rv != undefined)
                var mc: DependencyCondition = new ValueDependencyCondition(rv);
            else {
                let cond = d["cond"];
                if (cond != undefined) {
                    switch (cond) {
                        case "isvar":
                            var mc = new DependencyCondition(DependencyConditionType.IsVariable);
                            break;

                        case "isdisp":
                            var mc = new DependencyCondition(DependencyConditionType.IsDisplayed);
                            break;

                        default:
                            throw "Formulaire.parse_existence_dependencies() : type de condition '" + cond + "' non pris en charge";
                    }
                }
                else
                    throw "Formulaire.parse_existence_dependencies() : infos de dépendance manquantes/non prises en charge";
            }
            let dep = new ExistenceDependency(this, d["refid"], mc);
            this._dependencies.push(dep);
        }
    }

    /**
     * analyse les dépendances (existence/valeur)
     * @param json configuration de la dépendance
     * @param parentNode node parent dans lequel rechercher l'élément master (dont l'existence dépend)
     */
    public parseDependencies(json: {}) {
        const dep = json["dep_exist"];
        if (dep != undefined)
            this.parse_existence_dependencies(dep);
    }

    public verifiesDependency(d: Dependency): boolean {
        if (d.masterCondition.type == DependencyConditionType.IsDisplayed)
            return this._isDisplayed;
    }

    /**
     * applique une dépendance dont la condition a été vérifiée
     */
    public applyDependency(d: Dependency) {
        if (d instanceof ExistenceDependency)
            this.isDisplayed = true;
    }

    private prepareExistenceDependencies() {
        // si des FormulaireElement sont présents dans des ExistenceDependency, on met leur membre isDisplayed à false

        for (let d of this._dependencies)
            if (d instanceof ExistenceDependency) {
                this.isDisplayed = false;
                break;
            }
    }

    public applyDependencies(parentForm: FormulaireDefinition) {
        this.prepareExistenceDependencies();

        for (let d of this._dependencies) {
            d.apply();
        }

        for (const k of this.getKids())
            k.applyDependencies(parentForm);
    }

    public printDependencies() {
        for (let d of this._dependencies)
            console.log(d.toString());
    }


    public updateLocalisation(loc: StringMap, key?: string) {
        if (!key)
            key = this._confId;
        if (loc)
            this._label = loc[key];
        if (this._label == undefined)
            console.log(`WARNING : pas de traduction pour l'id ${this.id}`);

        for (let f of this.getKids())
            f.updateLocalisation(loc);
    }

    public toString() {
        return "id:" + this.id + (this._isDisplayed ? " displayed" : " NOT displayed") + " label:" + this.label;
    }

    public static removePrefix(s: string, prefix: string): string {
        if (s.startsWith(prefix)) {
            let l = prefix.length;
            return s.substr(l, s.length - l);
        }
        return undefined;
    }

    /**
     * itère sur tous les FormulaireElement
     */
    public get allFormElements(): IterableIterator<FormulaireElement> {
        return new DeepFormulaireElementIterator(this);
    }
}
