import { Injectable } from "@angular/core";
import { Response } from "@angular/http";
import { Observable as rxObservable } from "rxjs/Observable";
import "rxjs/add/operator/toPromise";
import { decode } from "he";
import { saveAs } from "file-saver";

import { CalculatorType, EnumEx, Observable } from "jalhyd";

import { ServiceFactory } from "../service-factory";
import { HttpService } from "../../services/http/http.service";
import { InternationalisationService } from "../../services/internationalisation/internationalisation.service";
import { FormulaireDefinition } from "../../formulaire/definition/form-definition";
import { FormulaireElement } from "../../formulaire/formulaire-element";
import { InputField } from "../../formulaire/input-field";
import { SelectField } from "../../formulaire/select-field";
import { CheckField } from "../../formulaire/check-field";
import { StringMap } from "../../stringmap";
import { FormulaireConduiteDistributrice } from "../../formulaire/definition/concrete/form-cond-distri";
import { FormulaireLechaptCalmon } from "../../formulaire/definition/concrete/form-lechapt-calmon";
import { FormulaireSectionParametree } from "../../formulaire/definition/concrete/form-section-parametree";
import { FormulaireCourbeRemous } from "../../formulaire/definition/concrete/form-courbe-remous";
import { FormulaireRegimeUniforme } from "../../formulaire/definition/concrete/form-regime-uniforme";
import { FormulairePasseBassinDimensions } from "../../formulaire/definition/concrete/form-passe-bassin-dim";
import { FormulairePasseBassinPuissance } from "../../formulaire/definition/concrete/form-passe-bassin-puissance";
import { FormulaireParallelStructure } from "../../formulaire/definition/concrete/form-parallel-structures";
import { FormulaireDever } from "../../formulaire/definition/concrete/form-dever";

@Injectable()
export class FormulaireService extends Observable {
    private _formulaires: FormulaireDefinition[];

    private _currentFormId = -1;

    constructor() {
        super();
        this._formulaires = [];
    }

    private get _intlService(): InternationalisationService {
        return ServiceFactory.instance.internationalisationService;
    }

    private get _httpService(): HttpService {
        return ServiceFactory.instance.httpService;
    }

    public get formulaires(): FormulaireDefinition[] {
        return this._formulaires;
    }

    private loadLocalisation(calc: CalculatorType): Promise<any> {
        let f: string = this.getConfigPathPrefix(calc) + this._intlService.currentLanguage.tag + ".json"
        let resp: rxObservable<Response> = this._httpService.httpGetRequestResponse(undefined, undefined, undefined, f);

        let prom = resp.map(res => res.text()).toPromise();
        return prom.then((res) => {
            let j = JSON.parse(res);
            return j as StringMap;
        })
    }

    /**
     * met à jour la langue du formulaire
     * @param formId id unique du formulaire
     * @param localisation ensemble id-message traduit
     */
    private updateFormulaireLocalisation(formId: number, localisation: StringMap) {
        for (let f of this._formulaires)
            if (f.uid == formId) {
                f.updateLocalisation(localisation);
                break;
            }
    }

    /**
     * met à jour l'aide du formulaire
     * @param formId id unique du formulaire
     */
    private updateFormulaireHelp(formId: number) {
        for (let f of this._formulaires)
            if (f.uid == formId) {
                f.updateHelp();
                break;
            }
    }

    /**
     * charge la localisation et met à jour la langue du formulaire
     */
    private loadUpdateFormulaireLocalisation(f: FormulaireDefinition): Promise<any> {
        return this.loadLocalisation(f.calculatorType)
            .then(localisation => {
                this.updateFormulaireLocalisation(f.uid, localisation);
                this.updateFormulaireHelp(f.uid);
            });
    }

    public updateLocalisation() {
        for (const c of EnumEx.getValues(CalculatorType)) {
            const prom: Promise<StringMap> = this.loadLocalisation(c);
            prom.then(loc => {
                for (const f of this._formulaires)
                    if (f.calculatorType == c)
                        this.updateFormulaireLocalisation(f.uid, loc);
            }
            );
        }
    }

    public getLocalisedTitleFromCalculatorType(type: CalculatorType) {
        const sCalculator: string = CalculatorType[type].toUpperCase();
        return this._intlService.localizeText(`INFO_${sCalculator}_TITRE`);
    }

    private loadConfig(form: FormulaireDefinition, ct: CalculatorType): Promise<Response> {
        let processData = function (s: string) {
            form.parseConfig(JSON.parse(s));
        }

        let f: string = this.getConfigPathPrefix(ct) + "config.json"
        return this._httpService.httpGetRequest(undefined, undefined, undefined, f, processData);
    }

    private newFormulaire(ct: CalculatorType, jsonState?: {}): FormulaireDefinition {
        let f: FormulaireDefinition;
        switch (ct) {
            case CalculatorType.ConduiteDistributrice:
                f = new FormulaireConduiteDistributrice();
                break;

            case CalculatorType.LechaptCalmon:
                f = new FormulaireLechaptCalmon();
                break;

            case CalculatorType.SectionParametree:
                f = new FormulaireSectionParametree();
                break;

            case CalculatorType.RegimeUniforme:
                f = new FormulaireRegimeUniforme();
                break;

            case CalculatorType.CourbeRemous:
                f = new FormulaireCourbeRemous();
                break;

            case CalculatorType.PabDimensions:
                f = new FormulairePasseBassinDimensions();
                break;

            case CalculatorType.PabPuissance:
                f = new FormulairePasseBassinPuissance();
                break;

            case CalculatorType.ParallelStructure:
                f = new FormulaireParallelStructure();
                break;

            case CalculatorType.Dever:
                f = new FormulaireDever();
                break;

            default:
                throw new Error(`FormulaireService.createFormulaire() : type de calculette ${ct} non pris en charge`)
        }

        if (jsonState !== undefined) {
            const props = jsonState["props"];
            f.initSessionNub(props);
        }
        else
            f.initSessionNub();

        return f;
    }

    /**
     * crée un formulaire d'un type donné
     * @param ct type de formulaire
     * @param jsonState
     */
    public createFormulaire(ct: CalculatorType, jsonState?: {}): Promise<FormulaireDefinition> {
        const f: FormulaireDefinition = this.newFormulaire(ct, jsonState);

        if (jsonState === undefined)
            f.calculatorName = decode(this.getLocalisedTitleFromCalculatorType(ct) + " (" + f.uid + ")");
        this._formulaires.push(f);
        let prom: Promise<Response> = this.loadConfig(f, ct);
        return prom.then(_ => {
            if (jsonState !== undefined)
                f.deserialiseJSON(jsonState);
            return f;
        }).then(f => {
            this.loadUpdateFormulaireLocalisation(f);
            return f;
        }).then(f => {
            f.applyDependencies();
            return f;
        }).then(f => {
            this.notifyObservers(
                {
                    "action": "createForm",
                    "form": f
                });
            return f;
        });
    }

    public getFormulaireFromId(uid: number): FormulaireDefinition {
        for (let f of this._formulaires)
            if (f.uid == uid)
                return f;

        return undefined;
    }

    public getInputField(formId: number, elemId: string): InputField {
        let s = this.getFormulaireElementById(formId, elemId);
        if (!(s instanceof InputField))
            throw "Form element with id '" + elemId + "' is not an input";
        return <InputField>s;
    }

    public getCheckField(formId: number, elemId: string): CheckField {
        let s = this.getFormulaireElementById(formId, elemId);
        if (!(s instanceof CheckField))
            throw "Form element with id '" + elemId + "' is not a checkbox";
        return <CheckField>s;
    }

    public getSelectField(formId: number, elemId: string): SelectField {
        let s = this.getFormulaireElementById(formId, elemId);
        if (!(s instanceof SelectField))
            throw "Form element with id '" + elemId + "' is not a select";
        return <SelectField>s;
    }

    private getFormulaireElementById(formId: number, elemId: string): FormulaireElement {
        for (let f of this._formulaires)
            if (f.uid == formId) {
                let s = f.getFormulaireNodeById(elemId);
                if (s != undefined)
                    return s as FormulaireElement;
            }
        return undefined;
    }

    public getConfigPathPrefix(ct: CalculatorType): string {
        if (ct == undefined)
            throw "FormulaireService.getConfigPathPrefix() : invalid undefined CalculatorType"

        switch (ct) {
            case CalculatorType.ConduiteDistributrice:
                return "app/calculators/cond_distri/cond_distri.";

            case CalculatorType.LechaptCalmon:
                return "app/calculators/lechapt-calmon/lechapt-calmon.";

            case CalculatorType.SectionParametree:
                return "app/calculators/section-param/section-param.";

            case CalculatorType.RegimeUniforme:
                return "app/calculators/regime-uniforme/regime-uniforme.";

            case CalculatorType.CourbeRemous:
                return "app/calculators/remous/remous.";

            case CalculatorType.PabDimensions:
                return "app/calculators/pab-dimensions/pab-dimensions.";

            case CalculatorType.PabPuissance:
                return "app/calculators/pab-puissance/pab-puissance.";

            case CalculatorType.Structure:
                return "app/calculators/ouvrages/ouvrages.";

            case CalculatorType.ParallelStructure:
                return "app/calculators/parallel-structures/parallel-structures.";

            case CalculatorType.Dever:
                return "app/calculators/dever/dever.";

            default:
                throw "FormulaireService.getConfigPathPrefix() : valeur de CalculatorType " + ct + " non implémentée"
        }
    }

    public requestCloseForm(uid: number) {
        const form = this.getFormulaireFromId(uid);
        if (form != undefined) {
            this._formulaires = this._formulaires.filter(f => f.uid !== uid);

            this.notifyObservers({
                "action": "closeForm",
                "form": form
            });
        }
    }

    public get currentFormId() {
        return this._currentFormId;
    }

    public setCurrentForm(formId: number) {
        const form = this.getFormulaireFromId(formId);
        if (form == undefined) {
            this._currentFormId = -1;
            this.notifyObservers({
                "action": "invalidFormId",
                "formId": formId
            });
        }
        else {
            this._currentFormId = formId;
            this.notifyObservers({
                "action": "currentFormChanged",
                "formId": formId
            });
        }
    }

    private get currentForm(): FormulaireDefinition {
        return this.getFormulaireFromId(this._currentFormId);
    }

    public get currentFormHasResults(): boolean {
        const form = this.currentForm;
        if (form != undefined)
            return form.hasResults;

        return false;
    }

    private readSingleFile(file: File): Promise<any> {
        return new Promise<any>((resolve, reject) => {
            var fr = new FileReader();

            fr.onload = () => {
                resolve(fr.result);
            };

            fr.onerror = () => {
                fr.abort();
                reject(new Error(`Erreur de lecture du fichier ${file.name}`));
            };

            fr.readAsText(file);
        });
    }

    /**
     * charge une session en tenant compte des calculettes sélectionnées
     * @param f fichier session
     * @param formInfos infos sur les calculettes @see LoadCalculatorComponent._calculators
     */
    public loadSession(f: File, formInfos: any[]) {
        this.readSingleFile(f).then(s => {
            const session = JSON.parse(s);
            for (const k in session)
                switch (k) {
                    case "session":
                        this.deserialiseSession(session[k], formInfos);
                        break;

                    default:
                        throw new Error(`session file : invalid key '${k}'`);
                }
        }).catch(err => {
            throw err;
        });
    }

    public saveSession(session: {}, filename?: string) {
        const blob = new Blob([JSON.stringify(session)], { type: "text/plain;charset=utf-8" });
        saveAs(blob, filename);
    }

    /**
     * obtient des infos (nom, uid des calculettes) d'un fichier session
     * @param f fichier session
     */
    public calculatorInfosFromSessionFile(f: File): Promise<any[]> {
        return this.readSingleFile(f).then(s => {
            const res: any[] = [];
            const session = JSON.parse(s);

            // liste des noms de calculettes
            for (const k in session)
                switch (k) {
                    case "session":
                        const sess = session[k];
                        const elems = sess["elements"];
                        for (const e of elems)
                            for (const k in e)
                                if (k === "form") {
                                    const form = e[k];
                                    res.push({ "uid": form["uid"], "title": form["id"] });
                                }
                        break;

                    default:
                        throw new Error(`session file : invalid key '${k}'`);
                }
            return res;
        });
    }

    public saveForm(f: FormulaireDefinition) {
        this.notifyObservers({
            "action": "saveForm",
            "form": f
        });
    }

    private deserialiseForm(elements: {}) {
        const props = elements["props"];
        const ct: CalculatorType = props["calcType"];
        this.createFormulaire(ct, elements);
    }

    private deserialiseSessionElement(element: {}, formInfos: any[]) {
        const keys = Object.keys(element);
        if (keys.length !== 1)
            throw new Error(`session file : invalid session object '${element}'`);

        switch (keys[0]) {
            case "form":
                const form = element[keys[0]];

                for (const i of formInfos)
                    if (i["uid"] == form["uid"] && i["selected"])
                        this.deserialiseForm(form);
                break;

            default:
                throw new Error(`session file : invalid key '${keys[0]}' in session object`);
        }
    }

    private deserialiseSession(elements: {}, formInfos: any[]) {
        for (const ks in elements)
            switch (ks) {
                case "elements":
                    for (const e of elements[ks])
                        this.deserialiseSessionElement(e, formInfos);
                    break;

                default:
                    throw new Error(`session file : invalid key '${ks}' in session object`);
            }
    }
}
